@@ -2,9 +2,9 @@
 
 __title__ = 'MK312'
 __description__ = 'MK312 Communication Wrapper'
-__url__ = ''
-__version__ = '0.2.0'
+__url__ = 'https://github.com/Rubberfate/mk312com'
+__version__ = '0.3.0'
 __author__ = 'Rubberfate'
 __author_email__ = 'rubberfate@gmail.com'
 __license__ = 'MIT'
-__copyright__ = 'Copyright 2020 Rubberfate'
+__copyright__ = 'Copyright 2021 Rubberfate'
